/*

remcmd.c - ZMON Remote Command mode tool

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#if defined(UNIX)
#include <termios.h>
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#if !defined(O_BINARY)
#define	O_BINARY 0
#endif
#include <time.h>
#include <errno.h>
#if defined(WIN32)
#include <windows.h>
#endif
/*...e*/

/*...smin:0:*/
#ifndef min
#define min(a,b) (((a)<(b))?(a):(b))
#endif
/*...e*/

/*...sfatal:0:*/
static void fatal(const char *fmt, ...)
	{
	va_list	vars;
	char s[256+1];
	va_start(vars, fmt);
	vsprintf(s, fmt, vars);
	va_end(vars);
	fprintf(stderr, "remcmd: %s\n", s);
	exit(1);
	}
/*...e*/

/*...sdelay_ms:0:*/
/* Delay upto 999ms */

static void delay_ms(int ms)
	{
#if defined(WIN32)
	Sleep(ms);
#elif defined(UNIX)
	struct timespec tv;
	tv.tv_sec  = 0;
	tv.tv_nsec = ms * 1000000;
	while ( nanosleep(&tv, &tv) == -1 && errno == EINTR )
		;
#else
  #error No suitable delay_ms function
#endif
	}
/*...e*/
/*...stty_select\44\ tty_finish:0:*/
static const char *name_tty = NULL;

#if defined(WIN32)

#define	DEF_SERIAL_DEVICE "COM1"

static HANDLE hSerial = INVALID_HANDLE_VALUE;

static void tty_select(const char *tty)
	{
	DCB dcb;
	if ( hSerial != INVALID_HANDLE_VALUE )
		{
		if ( !strcmp(tty, name_tty) )
			return;
		CloseHandle(hSerial);
		hSerial = INVALID_HANDLE_VALUE;
		}

	if ( (hSerial = CreateFile(
		tty,
		GENERIC_READ|GENERIC_WRITE,
		0,				/* Don't share */
		(LPSECURITY_ATTRIBUTES) NULL,
		OPEN_EXISTING,
		0,
		(HANDLE) NULL)) == INVALID_HANDLE_VALUE )
		fatal("can't open tty %s", tty);

	memset(&dcb, 0, sizeof(dcb));
	dcb.DCBlength         = sizeof(DCB);
	dcb.BaudRate          = 19200;
	dcb.fBinary           = TRUE;
	dcb.fParity           = TRUE;
	dcb.fOutxCtsFlow      = FALSE; /* Always send, regardless of CTS */
	dcb.fOutxDsrFlow      = FALSE; /* Always send, regardless of DSR */
	dcb.fDtrControl       = DTR_CONTROL_ENABLE; /* We're ready to receive */
	dcb.fDsrSensitivity   = FALSE; /* Ignore DSR from modem */
	dcb.fOutX             = FALSE; /* No XON/XOFF flow control on output */
	dcb.fInX              = FALSE; /* No XON/XOFF flow control on input */
	dcb.fRtsControl       = RTS_CONTROL_ENABLE; /* We're ready to receive */
	dcb.fAbortOnError     = FALSE;
	dcb.ByteSize          = 8; /* 4, 5, 6, 7, 8 */
	dcb.Parity            = NOPARITY;
	dcb.StopBits          = TWOSTOPBITS;
	dcb.XonChar           = 0; /* Any known value */
	dcb.XoffChar          = 1; /* Must not equal XonChar */
	SetCommState(hSerial, &dcb);
	}

static void tty_finish(void)
	{
	if ( hSerial != INVALID_HANDLE_VALUE )
		{
		CloseHandle(hSerial);
		hSerial = INVALID_HANDLE_VALUE;
		}
	}

static void tty_send(unsigned char b)
	{
	DWORD cbWritten;
	WriteFile(hSerial, &b, 1, &cbWritten, (LPOVERLAPPED)NULL);
	}

static unsigned char tty_recv(void)
	{
	unsigned char b;
	DWORD cbRead;
	ReadFile(hSerial, &b, 1, &cbRead, (LPOVERLAPPED)NULL);
	return b;
	}

#elif defined(LINUX)

#define DEF_SERIAL_DEVICE "/dev/ttyS0"

static int fd_tty = -1;

static void tty_select(const char *tty)
	{
	struct termios ta;
	if ( fd_tty != -1 )
		{
		if ( !strcmp(tty, name_tty) )
			return;
		close(fd_tty);
		fd_tty = -1;
		}
	if ( (fd_tty = open(tty, O_RDWR|O_BINARY)) == -1 )
		fatal("can't open tty %s", tty);
	tcgetattr(fd_tty, &ta);
	ta.c_iflag = ( IGNBRK | IGNPAR ); /* ignore break and parity */
	ta.c_oflag = 0;
	ta.c_cflag = CS8 | CSTOPB | CLOCAL | CREAD; /* 8 bit, 2 stop bits, no parity, no modem control lines, no hardware flow control */
	ta.c_lflag = 0;
	cfsetospeed(&ta, B19200);
	cfsetispeed(&ta, B19200);
	tcsetattr(fd_tty, TCSANOW, &ta);
	name_tty = tty;
	}

static void tty_finish(void)
	{
	if ( fd_tty != -1 )
		{
		close(fd_tty);
		fd_tty = -1;
		}
	}

static void tty_send(unsigned char b)
	{
	write(fd_tty, &b, 1);
	}

static unsigned char tty_recv(void)
	{
	unsigned char b;
	read(fd_tty, &b, 1);
	return b;
	}

#else
  #error No suitable tty_ routines
#endif
/*...e*/

/*...susage:0:*/
static void usage(void)
	{
	fprintf(stderr, "usage: remcmd {flags}\n");
	fprintf(stderr, "flags: -t tty     set tty (default %s)\n", DEF_SERIAL_DEVICE);
	fprintf(stderr, "       -p port    set port (default 0x00)\n");
	fprintf(stderr, "       -v value   set value (default 0x00)\n");
	fprintf(stderr, "       -s addr    set MTX src address (default 0x0100)\n");
	fprintf(stderr, "       -d addr    set MTX dst address (default 0x0100)\n");
	fprintf(stderr, "       -x addr    set MTX exe address (default 0x0100)\n");
	fprintf(stderr, "       -g d:t     set drive letter:config (default B:03)\n");
	fprintf(stderr, "       -f fn      set local filename (default remcmd.dat)\n");
	fprintf(stderr, "       -l length  set length (default 0x0000)\n");
	fprintf(stderr, "       -c         toggle compressed read/write mode (default uncompressed)\n");
	fprintf(stderr, "       -F         send file (good for testing connectivity)\n");
	fprintf(stderr, "       -X         send to MTX dst in Intel Hex and jump to MTX exe address\n");
	fprintf(stderr, "       -I         input from port\n");
	fprintf(stderr, "       -O         output value to port\n");
	fprintf(stderr, "       -R         read length bytes from MTX src address and write to file\n");
	fprintf(stderr, "       -W         read bytes from file and write to MTX dst address\n");
	fprintf(stderr, "       -G         read length bytes from MTX RAM Disc and write to file\n");
	fprintf(stderr, "       -P         read bytes from file and write to MTX RAM Disc\n");
	fprintf(stderr, "       -D         read length bytes from MTX Disc and write to file\n");
	fprintf(stderr, "       -S         read bytes from file and write to MTX Disc\n");
	fprintf(stderr, "       -C         copy length bytes from MTX src to MTX dst\n");
	fprintf(stderr, "       -J         jump to MTX exe address\n");
	fprintf(stderr, "       -K         call MTX exe address\n");
	exit(1);
	}
/*...e*/

/*...sdo_in:0:*/
static unsigned char do_in(unsigned char port)
	{
	tty_send(0x01);
	tty_send(port);
	return tty_recv();
	}
/*...e*/
/*...sdo_out:0:*/
static void do_out(unsigned char port, unsigned char value)
	{
	tty_send(0x02);
	tty_send(port);
	tty_send(value);
	tty_recv();
	}
/*...e*/
/*...sdo_read:0:*/
static void do_read(unsigned src_addr, unsigned char *buf, unsigned length, int compressed)
	{
	if ( compressed )
		{
		tty_send(0x05);
		tty_send(src_addr&0xff);
		tty_send(src_addr>>8);
		tty_send(length);
		while ( length > 0 )
			{
			signed char l = tty_recv();
			if ( l >= 0 )
				{
				signed char i;
				for ( i = 0; i < l; i++ )
					*buf++ = tty_recv();
				}
			else
				{
				unsigned char b = tty_recv();
				signed char i;
				l = -l;
				for ( i = 0; i < l; i++ )
					*buf++ = b;
				}
			length -= l;
			}		
		}
	else
		{
		unsigned i;
		tty_send(0x03);
		tty_send(src_addr&0xff);
		tty_send(src_addr>>8);
		tty_send(length);
		for ( i = 0; i < length; i++ )
			buf[i] = tty_recv();
		}
	}
/*...e*/
/*...sdo_write:0:*/
/*...srun_length:0:*/
static unsigned run_length(const unsigned char *buf, unsigned length)
	{
	unsigned char b = buf[0];
	unsigned l = 1;
	while ( l < length && l < 127 && b == buf[l] )
		++l;
	return l;
	}
/*...e*/
/*...slit_length:0:*/
static unsigned lit_length(const unsigned char *buf, unsigned length)
	{
	unsigned l = 1;
	while ( l < length && l < 127 && buf[l-1] != buf[l] )
		++l;
	return l;
	}
/*...e*/

static void do_write(const unsigned char *buf, unsigned dst_addr, unsigned length, int compressed)
	{
	unsigned i;
	if ( compressed )
		{
		tty_send(0x06);
		tty_send(dst_addr&0xff);
		tty_send(dst_addr>>8);
		while ( length > 0 )
			{
			unsigned l = run_length(buf, length);
			if ( l > 1 )
				{
				tty_send((unsigned char)(-(signed char)l));
				tty_send(buf[0]);
				buf += l;
				}
			else
				{
				l = lit_length(buf, length);
				tty_send(l);
				for ( i = 0; i < l; i++ )
					tty_send(*buf++);
				}
			length -= l;
			}
		tty_send(0x00);
		}
	else
		{
		tty_send(0x04);
		tty_send(dst_addr&0xff);
		tty_send(dst_addr>>8);
		tty_send(length);
		for ( i = 0; i < length; i++ )
			tty_send(buf[i]);
		}
	tty_recv();
	}

static void do_write_byte(unsigned char b, unsigned dst_addr)
	{
	do_write(&b, dst_addr, 1, 0);
	}

static void do_write_word(unsigned short w, unsigned dst_addr)
	{
	unsigned char b[2];
	b[0] = (unsigned char) (w&0xff);
	b[1] = (unsigned char) (w>>8);
	do_write(b, dst_addr, 2, 0);
	}
/*...e*/
/*...sdo_jump:0:*/
static void do_jump(unsigned exe_addr)
	{
	tty_send(0x08);
	tty_send(exe_addr&0xff);
	tty_send(exe_addr>>8);
	}
/*...e*/
/*...sdo_call:0:*/
static unsigned char do_call(unsigned exe_addr)
	{
	tty_send(0x08);
	tty_send(exe_addr&0xff);
	tty_send(exe_addr>>8);
	return tty_recv();
	}
/*...e*/
/*...sdisc constants:0:*/
#define	SECBUF	0xdf00		/* 128 bytes */
#define	DRVRQ	0xffe8		/* byte, 0..7 for B:..I: */
#define	CFGBYT	0xffe9		/* byte */
#define	TRKRQ	0xffea		/* word, 0.. */
#define	SECRQ	0xffec		/* word, 1..26 */
#define	DMARQ	0xffee		/* word */
#define	EXCNFG	0xfff0		/* call */
#define	EXRD	0xfff3		/* call */
#define	EXWR	0xfff6		/* call */
#define	INITLZ	0xfffc		/* call */
/*...e*/
/*...sdo_config:0:*/
static void do_config(unsigned char drive, unsigned char config)
	{
	do_call(INITLZ);
	do_write_byte(drive-'B', DRVRQ);
	do_write_byte(config, CFGBYT);
	if ( do_call(EXCNFG) != 0 )
		fatal("can't configure drive %c:%02x", drive, config); 
	do_write_word(SECBUF, DMARQ);
	}
/*...e*/
/*...sdo_read_sector:0:*/
static unsigned char do_read_sector(unsigned track, unsigned sector, unsigned char *buf, int compressed)
	{
	unsigned char a;
	do_write_word(track, TRKRQ);
	do_write_word(sector, SECRQ);
	a = do_call(EXRD);
	if ( a == 0 )
		do_read(SECBUF, buf, 128, compressed);
	return a;
	}

/*...e*/
/*...sdo_write_sector:0:*/
static unsigned char do_write_sector(const unsigned char *buf, unsigned track, unsigned sector, int compressed)
	{
	do_write(buf, SECBUF, 128, compressed);
	do_write_word(track, TRKRQ);
	do_write_word(sector, SECRQ);
	return do_call(EXWR);
	}
/*...e*/

/*...scmd_file:0:*/
static void cmd_file(const char *fn)
	{
	FILE *fp;
	int c;
	if ( (fp = fopen(fn, "rb")) == NULL )
		fatal("can't open %s", fn);
	while ( (c = fgetc(fp)) != EOF )
		{
		fputc(c, stdout);
		fflush(stdout);
		tty_send((unsigned char) c);
		if ( c == '\n' )
			delay_ms(200);
		}
	fclose(fp);
	printf("send(%s)\n", fn);
	}
/*...e*/
/*...scmd_intel_hex_file:0:*/
/*...ssend_hex:0:*/
static void send_hex(unsigned char b)
	{
	static char hexdigits[] = "0123456789ABCDEF";
	tty_send(hexdigits[b>>4]);
	tty_send(hexdigits[b&15]);
	}
/*...e*/
/*...ssend_record:0:*/
static void send_record(
	unsigned char record_type,
	unsigned addr,
	size_t len,
	const unsigned char *buf,
	unsigned record_num
	)
	{
	unsigned char record_len = (unsigned char) len;
	unsigned char addr_hi    = (unsigned char) (addr>>8);
	unsigned char addr_lo    = (unsigned char)  addr    ;
	unsigned char cs         = 0;
	size_t i;
	tty_send(':');
	send_hex(record_len); cs += record_len;
	/* ZMON outputs a '.' for every record at this point,
	   but we've made it switch to page mode
	   to avoid needing the following hack. */
#if 0
	if ( record_num >= 20*56 && (record_num%56) == 0 )
		/* SCPM CRT implementation scrolls very slowly,
		   but we've attempted to minimise this by selecting
		   the screen mode that doesn't update colours and
		   clearing the screen so there are 20 lines that
		   can be output before any scrolling occurs.
		   At 32 bytes per record, that's over 35KB. */
		{
		tcdrain(fd_tty);
		delay_ms(200);
		}
#endif
	send_hex(addr_hi); cs += addr_hi;
	send_hex(addr_lo); cs += addr_lo;
	/* We send the correct record_type, but ZMON doesn't check it. */
	send_hex(record_type); cs += record_type;
	for ( i = 0; i < len; i++ )
		{ send_hex(buf[i]); cs += buf[i]; }
	/* We send the correct checksum, but ZMON doesn't check it. */
	send_hex(-cs);
	/* Technically the record is supposed to be terminated by newline
	   character(s), but ZMON doesn't need these, so don't send. */
	}
/*...e*/

static void cmd_intel_hex_file(const char *fn, unsigned dst_addr, unsigned exe_addr)
	{
	FILE *fp;
	unsigned char buf[32];
	size_t n;
	unsigned record_num = 0;
	if ( (fp = fopen(fn, "rb")) == NULL )
		fatal("can't open %s for reading", fn);
	while ( (n = fread(buf, 1, 32, fp)) > 0 )
		{
		send_record(0x00, dst_addr, n, buf, ++record_num);
		dst_addr += n;
		}
	fclose(fp);
	send_record(0x01, exe_addr, 0, NULL, 0);
	printf("send_intel_hex(%s,0x%04x,0x%04x)\n", fn, dst_addr, exe_addr);
	}
/*...e*/
/*...scmd_in:0:*/
static void cmd_in(unsigned char port)
	{
	unsigned char value = do_in(port);
	printf("in(0x%02x) -> 0x%02x\n", (unsigned) port, (unsigned) value);
	}
/*...e*/
/*...scmd_out:0:*/
static void cmd_out(unsigned char port, unsigned char value)
	{
	tty_send(0x02);
	tty_send(port);
	tty_send(value);
	tty_recv();
	printf("out(0x%02x,0x%02x)\n", (unsigned) port, (unsigned) value);
	}
/*...e*/
/*...scmd_read:0:*/
static void cmd_read(unsigned src_addr, unsigned length, const char *fn, int compressed)
	{
	unsigned addr = src_addr;
	FILE *fp;
	if ( (fp = fopen(fn, "wb")) == NULL )
		fatal("can't open %s for writing", fn);
	while ( length > 0 )
		{
		unsigned thisgo = min(length, 255);
		unsigned char buf[255];
		do_read(addr, buf, thisgo, compressed);
		fwrite(buf, 1, thisgo, fp);
		addr   += thisgo;
		length -= thisgo;
		}
	fclose(fp);
	printf("read(0x%04x,%s)\n", src_addr, fn);
	}
/*...e*/
/*...scmd_write:0:*/
static void cmd_write(const char *fn, unsigned dst_addr, int compressed)
	{
	FILE *fp;
	unsigned char buf[255];
	unsigned addr = dst_addr;
	size_t thisgo;
	if ( (fp = fopen(fn, "rb")) == NULL )
		fatal("can't open %s for reading", fn);
	while ( (thisgo = fread(buf, 1, 255, fp)) > 0 )
		{
		do_write(buf, addr, thisgo, compressed);
		addr += thisgo;
		}
	printf("write(%s,0x%04x)\n", fn, dst_addr);
	}
/*...e*/
/*...scmd_read_ramdisc:0:*/
static void cmd_read_ramdisc(unsigned length, const char *fn, int compressed)
	{
	unsigned char pg = 0x81;
	FILE *fp;
	if ( (fp = fopen(fn, "wb")) == NULL )
		fatal("can't open %s for writing", fn);
	while ( length > 0 )
		{
		unsigned length_pg = min(length, 0xc000);
		unsigned addr = 0x0000;
		do_out(0x00, pg++);
		while ( length_pg > 0 )
			{
			unsigned thisgo = min(length_pg, 255);
			unsigned char buf[255];
			do_read(addr, buf, thisgo, compressed);
			fwrite(buf, 1, thisgo, fp);
			addr      += thisgo;
			length_pg -= thisgo;
			length    -= thisgo;
			}
		}
	fclose(fp);
	do_out(0x00, 0x80);
	printf("read_ramdisc(%s)\n", fn);
	}
/*...e*/
/*...scmd_write_ramdisc:0:*/
static void cmd_write_ramdisc(const char *fn, int compressed)
	{
	unsigned char pg = 0x81;
	unsigned char buf[0xc000];
	size_t length_pg;
	FILE *fp;
	if ( (fp = fopen(fn, "rb")) == NULL )
		fatal("can't open %s for reading", fn);
	while ( (length_pg = fread(buf, 1, 0xc000, fp)) > 0 )
		{
		unsigned addr = 0x0000;
		do_out(0x00, pg++);
		while ( length_pg > 0 )
			{
			unsigned thisgo = min(length_pg, 255);
			do_write(buf+addr, addr, thisgo, compressed);
			addr      += thisgo;
			length_pg -= thisgo;
			}
		}
	fclose(fp);
	do_out(0x00, 0x80);
	printf("write_ramdisc(%s)\n", fn);
	}
/*...e*/
/*...scmd_read_disc:0:*/
static void cmd_read_disc(
	unsigned char drive,
	unsigned char config,
	unsigned length,
	const char *fn,
	int compressed
	)
	{
	unsigned track = 0, sector = 1;
	FILE *fp;
	do_config(drive, config);
	if ( (fp = fopen(fn, "wb")) == NULL )
		fatal("can't open %s for writing", fn);
	while ( length >= 128 )
		{
		unsigned char buf[128];
		int i;
		for ( i = 0; i < 10; i++ )
 			if ( do_read_sector(track, sector, buf, compressed) == 0 )
 				break;
		if ( i == 10 )
			fatal("error reading disc, track=%u, sector=%u", track, sector);
		fwrite(buf, 1, 128, fp);
		if ( sector++ == 26 )
			{ sector = 1; track++; }
		length -= 128;
		}
	fclose(fp);
	printf("read_disc(%s)\n", fn);
	}
/*...e*/
/*...scmd_write_disc:0:*/
static void cmd_write_disc(const char *fn, unsigned char drive, unsigned char config, int compressed)
	{
	unsigned track = 0, sector = 1;
	FILE *fp;
	unsigned char buf[128];
	do_config(drive, config);
	if ( (fp = fopen(fn, "rb")) == NULL )
		fatal("can't open %s for reading", fn);
	while ( fread(buf, 1, 128, fp) == 128 )
		{
		int i;
		for ( i = 0; i < 10; i++ )
			if ( do_write_sector(buf, track, sector, compressed) == 0 )
				break;
		if ( i == 10 )
			fatal("error writing disc, track=%u, sector=%u", track, sector);
		if ( sector++ == 26 )
			{ sector = 1; track++; }
		}
	fclose(fp);
	printf("write_disc(%s)\n", fn);
	}
/*...e*/
/*...scmd_copy:0:*/
static void cmd_copy(unsigned src_addr, unsigned dst_addr, unsigned length)
	{
	if ( length > 0 )
		{
		tty_send(0x07);
		tty_send(src_addr&0xff);
		tty_send(src_addr>>8);
		tty_send(dst_addr&0xff);
		tty_send(dst_addr>>8);
		tty_send(length&0xff);
		tty_send(length>>8);
		tty_recv();
		}
	printf("copy(0x%04x,0x%04x,0x%04x)\n", src_addr, dst_addr, length);
	}
/*...e*/
/*...scmd_jump:0:*/
static void cmd_jump(unsigned exe_addr)
	{
	do_jump(exe_addr);
	printf("jump(0x%04x)\n", exe_addr);
	}
/*...e*/
/*...scmd_call:0:*/
static void cmd_call(unsigned exe_addr)
	{
	unsigned char a = do_call(exe_addr);
	printf("call(0x%04x) -> 0x%02x\n", exe_addr, (unsigned) a);
	}
/*...e*/

/*...smain:0:*/
int main(int argc, char *argv[])
	{
	const char *tty      = DEF_SERIAL_DEVICE;
	unsigned char port   = 0x00;
	unsigned char value  = 0x00;
	unsigned src_addr    = 0x0100;
	unsigned dst_addr    = 0x0100;
	unsigned exe_addr    = 0x0100;
	const char *fn       = "remcmd.dat";
	unsigned length      = 0x0000;
	unsigned char drive  = 'B';
	unsigned char config = 0x03;
	int compressed       = 0;
	int i, tmp;

	for ( i = 1; i < argc; i++ )
		{
		if ( argv[i][0] != '-' )
			break;
		else if ( argv[i][1] == '-' )
			{ ++i; break; }
		switch ( argv[i][1] )
			{
/*...st \45\ set tty:24:*/
case 't':
	if ( ++i == argc )
		usage();
	tty = argv[i];
	break;
/*...e*/
/*...sp \45\ set port:24:*/
case 'p':
	{
	if ( ++i == argc )
		usage();
	sscanf(argv[i], "%i", &tmp);
	port = (unsigned char) tmp;
	}
	break;
/*...e*/
/*...sv \45\ set value:24:*/
case 'v':
	{
	if ( ++i == argc )
		usage();
	sscanf(argv[i], "%i", &tmp);
	value = (unsigned char) tmp;
	}
	break;
/*...e*/
/*...ss \45\ set src_addr:24:*/
case 's':
	{
	if ( ++i == argc )
		usage();
	sscanf(argv[i], "%i", &tmp);
	src_addr = (unsigned) tmp;
	}
	break;
/*...e*/
/*...sd \45\ set dst_addr:24:*/
case 'd':
	{
	if ( ++i == argc )
		usage();
	sscanf(argv[i], "%i", &tmp);
	dst_addr = (unsigned) tmp;
	}
	break;
/*...e*/
/*...sx \45\ set exe_addr:24:*/
case 'x':
	{
	if ( ++i == argc )
		usage();
	sscanf(argv[i], "%i", &tmp);
	exe_addr = (unsigned) tmp;
	}
	break;
/*...e*/
/*...sg \45\ set drive\58\config:24:*/
case 'g':
	{
	if ( ++i == argc )
		usage();
	sscanf(argv[i], "%c:%02x", &drive, &tmp);
	if ( drive < 'B' || drive > 'I' )
		fatal("bad drive letter");
	config = (unsigned char) tmp;
	if (                    config <= 0x07  || /* 3.5" or 5.25" Disc */
	     (config >= 0x10 && config <= 0x13) || /* 8" Disc */
	     (config >= 0x18 && config <= 0x1f) || /* SD/CF Card */
	     (config >= 0x40 && config <= 0x4f) )  /* Silicon Disc */
		;
	else if ( config >= 0x50 && config <= 0x52 ) /* RAM Disc */
		fatal("use the RAM Disc -G and -P functions instead");
	else
		fatal("invalid drive config %c:%02x", (int) drive, config);
	}
	break;
/*...e*/
/*...sf \45\ set fn:24:*/
case 'f':
	if ( ++i == argc )
		usage();
	fn = argv[i];
	break;
/*...e*/
/*...sl \45\ set length:24:*/
case 'l':
	{
	if ( ++i == argc )
		usage();
	sscanf(argv[i], "%i", &tmp);
	length = (unsigned) tmp;
	}
	break;
/*...e*/
/*...sc \45\ toggle compress flag:24:*/
case 'c':
	compressed = !compressed;
	break;
/*...e*/
/*...sF \45\ send file:24:*/
case 'F':
	tty_select(tty);
	cmd_file(fn);
	break;
/*...e*/
/*...sX \45\ send file as intel hex:24:*/
case 'X':
	tty_select(tty);
	cmd_intel_hex_file(fn, dst_addr, exe_addr);
	break;
/*...e*/
/*...sI \45\ in:24:*/
case 'I':
	tty_select(tty);
	cmd_in(port);
	break;
/*...e*/
/*...sO \45\ out:24:*/
case 'O':
	tty_select(tty);
	cmd_out(port, value);
	break;
/*...e*/
/*...sR \45\ read:24:*/
case 'R':
	tty_select(tty);
	cmd_read(src_addr, length, fn, compressed);
	break;
/*...e*/
/*...sW \45\ write:24:*/
case 'W':
	tty_select(tty);
	cmd_write(fn, dst_addr, compressed);
	break;
/*...e*/
/*...sG \45\ read RAM Disc:24:*/
case 'G':
	tty_select(tty);
	cmd_read_ramdisc(length, fn, compressed);
	break;
/*...e*/
/*...sP \45\ write RAM Disc:24:*/
case 'P':
	tty_select(tty);
	cmd_write_ramdisc(fn, compressed);
	break;
/*...e*/
/*...sD \45\ read Disc:24:*/
case 'D':
	tty_select(tty);
	cmd_read_disc(drive, config, length, fn, compressed);
	break;
/*...e*/
/*...sS \45\ write Disc:24:*/
case 'S':
	tty_select(tty);
	cmd_write_disc(fn, drive, config, compressed);
	break;
/*...e*/
/*...sC \45\ copy:24:*/
case 'C':
	tty_select(tty);
	cmd_copy(src_addr, dst_addr, length);
	break;
/*...e*/
/*...sJ \45\ jump:24:*/
case 'J':
	tty_select(tty);
	cmd_jump(exe_addr);
	break;
/*...e*/
/*...sK \45\ call:24:*/
case 'K':
	tty_select(tty);
	cmd_call(exe_addr);
	break;
/*...e*/
/*...sdefault:24:*/
default:
	usage();
	break;
/*...e*/
			}
		}
	if ( i != argc )
		usage();

	tty_finish();
	return 0;
	}
/*...e*/
